#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import os

import pytest
import mock

import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.enums as enums
import octobot_trading.exchange_data.contracts as contracts
from octobot_trading.personal_data import FuturePortfolio, BuyMarketOrder, SellMarketOrder, SellLimitOrder, \
    StopLossOrder, LinearPosition, InversePosition, get_max_order_quantity_for_price

from tests import event_loop
from tests.exchanges import future_simulated_exchange_manager
from tests.exchanges.traders import future_trader, future_trader_simulator_with_default_linear, DEFAULT_FUTURE_SYMBOL, \
    DEFAULT_FUTURE_SYMBOL_MARGIN_TYPE, DEFAULT_FUTURE_SYMBOL_LEVERAGE, future_trader_simulator_with_default_inverse
from tests.test_utils.order_util import fill_market_order

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def _update_position_mark_price(exchange_manager_inst, mark_price, symbol=DEFAULT_FUTURE_SYMBOL, side=None):
    symbol_position = exchange_manager_inst.exchange_personal_data.positions_manager. \
        get_symbol_position(symbol=symbol, side=side)
    await symbol_position.update(mark_price=mark_price)


async def test_initial_portfolio(future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager

    assert isinstance(portfolio_manager.portfolio, FuturePortfolio)


async def test_update_future_portfolio_from_order(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(2))

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    buy_price = decimal.Decimal(str(100))
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=buy_price,
                      quantity=decimal.Decimal(str(2)),
                      price=buy_price)

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(900))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").initial_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").wallet_balance == decimal.Decimal(str(1000))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == decimal.Decimal(str(100))

    # create a new position from order
    await fill_market_order(market_buy)

    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(899.92))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(999.92))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == 0
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").initial_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").wallet_balance == decimal.Decimal(str(999.92))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal(str(100))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO


async def test_update_portfolio_available_from_order_with_market_buy_long_linear_contract(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(5))

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(str(1000)),
                      quantity=decimal.Decimal(str(2.5)),  # real quantity = 2.5 / 5 = 0.5 BTC at 1000$
                      price=decimal.Decimal(str(1000)))

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    # 1000 - 0.5 BTC * 1000$ = 500
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(500))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == decimal.Decimal(500)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(str(74)),
                      quantity=decimal.Decimal(str(5)),  # real quantity = 5 / 5 = 1 BTC at 74$
                      price=decimal.Decimal(str(74)))

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(1000)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == decimal.Decimal(574)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
    # 500 - 1 BTC * 74$ = 426
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(426))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO

    # test cancel buy order
    portfolio_manager.portfolio.update_portfolio_available(market_buy, False)
    # 426 + 1 BTC * 74$ = 500
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(500))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == decimal.Decimal(str(500))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(10)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(str(10))


async def test_update_portfolio_available_from_order_with_market_buy_long_inverse_contract(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    symbol_contract = contracts.FutureContract(
        pair=DEFAULT_FUTURE_SYMBOL,
        margin_type=DEFAULT_FUTURE_SYMBOL_MARGIN_TYPE,
        contract_type=enums.FutureContractType.INVERSE_PERPETUAL,
        current_leverage=DEFAULT_FUTURE_SYMBOL_LEVERAGE)
    trader_inst.exchange_manager.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, symbol_contract)

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(str(1000)),
                      quantity=decimal.Decimal(str(1000)),  # real quantity = 1000 / 1000 = 1 BTC at 1000$
                      price=decimal.Decimal(str(1000)))

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(str(9))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(str(10))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ONE
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(str(74)),
                      quantity=decimal.Decimal(str(74 / 3)),  # real quantity = 0.3 BTC at 74$
                      price=decimal.Decimal(str(74)))

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(10)
    # 9 - (74 / 3) / 74 = 8.666666666666666648648648649
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '8.666666666666666648648648649')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == decimal.Decimal(
        str('1.333333333333333351351351351'))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO

    # test cancel buy order
    portfolio_manager.portfolio.update_portfolio_available(market_buy, False)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(9)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(str(10))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == decimal.Decimal(
        str('0.9999999999999999999999999996'))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(1000))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000))


async def test_update_portfolio_data_from_order_with_market_buy_long_linear_contract(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(5))
    _disable_fees(exchange_manager_inst)

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    buy_price = decimal.Decimal(str(1000))
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=buy_price,
                      quantity=decimal.Decimal(3),  # real quantity = 3 / 5 = 0.6
                      price=buy_price)  # real cost = 0.6 * 1000 = 600

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)

    # create position from order
    await fill_market_order(market_buy)

    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(400.0))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000.0))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal(str(600.0))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=buy_price,
                      quantity=decimal.Decimal(1),  # 0.2
                      price=buy_price)  # 200

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    await fill_market_order(market_buy)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(200.0))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000.0))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal(str(800.0))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO

    # Test reducing LONG position with a sell market order
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=buy_price,
                       quantity=decimal.Decimal(0.5),  # 0.1
                       price=buy_price)  # 100

    # Should restore available
    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    await fill_market_order(market_sell)
    fees_1 = market_sell.get_total_fees(market_sell.market)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(10)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(300)) - fees_1
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(10)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000)) - fees_1
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal(str(700.0))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO

    # Test buy order 2
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=buy_price,
                      quantity=decimal.Decimal(0.5),  # 0.1
                      price=buy_price)  # 100

    # increase position size
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    await fill_market_order(market_buy)
    fees_2 = market_buy.get_total_fees(market_buy.market)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(200)) - fees_1 \
           - fees_2
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000)) - fees_1 \
           - fees_2
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal(str(800.0))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO

    # Test reducing again LONG position with a sell market order
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=buy_price,
                       quantity=decimal.Decimal(4),  # 0.8
                       price=buy_price)  # 800

    # decrease position size
    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    await fill_market_order(market_sell)
    fees_3 = market_sell.get_total_fees(market_sell.market)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(10)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(1000))- fees_1 \
           - fees_2 - fees_3
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(10)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000))- fees_1 \
           - fees_2 - fees_3
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO


async def test_update_portfolio_data_from_order_that_triggers_negative_portfolio_linear_contract(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(5))

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(str(1000)),
                      quantity=decimal.Decimal(str(100000000)),
                      price=decimal.Decimal(str(1000)))

    # test buy order creation
    # if not os.getenv('CYTHON_IGNORE'):
    with pytest.raises(errors.PortfolioNegativeValueError):
        portfolio_manager.portfolio.update_portfolio_available(market_buy, True)


async def test_update_portfolio_data_from_order_with_cancelled_and_filled_orders_linear_contract(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(5))

    # Test sell order
    market_sell = SellMarketOrder(trader_inst)
    sell_price = decimal.Decimal(str(80))
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=sell_price,
                       quantity=decimal.Decimal(str(12)),
                       price=sell_price)

    # Test sell order
    limit_sell = SellLimitOrder(trader_inst)
    limit_sell.update(order_type=enums.TraderOrderType.SELL_LIMIT,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(str(10)),
                      quantity=decimal.Decimal(str(46.5)),
                      price=decimal.Decimal(str(10)))

    # Test stop loss order
    stop_loss = StopLossOrder(trader_inst)
    stop_loss.update(order_type=enums.TraderOrderType.STOP_LOSS,
                     symbol=DEFAULT_FUTURE_SYMBOL,
                     current_price=decimal.Decimal(str(80)),
                     quantity=decimal.Decimal(str(46.5)),
                     price=decimal.Decimal(str(80)))

    portfolio_manager.portfolio.update_portfolio_available(stop_loss, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('907.0')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('715.0')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    await fill_market_order(market_sell)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('714.616')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('999.616')

    # cancel other orders
    portfolio_manager.portfolio.update_portfolio_available(stop_loss, False)
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, False)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('807.616')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('999.616')


async def test_update_portfolio_data_from_orders_with_max_long_size_linear_contract(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    leverage = decimal.Decimal(2)
    default_contract.set_current_leverage(leverage)

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    buy_order_price = decimal.Decimal(40)
    leveraged_max_buy_quantity = get_max_order_quantity_for_price(
        LinearPosition(trader_inst, default_contract),
        portfolio_manager.portfolio.get_currency_portfolio("USDT").available,
        buy_order_price,
        enums.PositionSide.LONG,
        DEFAULT_FUTURE_SYMBOL
    )
    assert 49 < leveraged_max_buy_quantity < 50   # with a x1 leverage, it would be 24 < max_buy_quantity < 25

    max_buy_quantity = leveraged_max_buy_quantity / leverage
    # WARNING:
    # Later on in this test, use max_buy_quantity only to avoid last digit rounding issues on later asserts
    # that are due to the previous line's division by leverage.
    # This has no impact on trading (it's on the 24th digit) but is failing "==" tests
    # rounding error example:
    #   AssertionError: assert Decimal('2992.017946161515453639082751') == Decimal('2992.017946161515453639082752')
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=buy_order_price,
                      quantity=max_buy_quantity * leverage,    # apply leverage (restore leveraged_max_buy_quantity)
                      price=buy_order_price,
                      fees_currency_side=default_contract.get_fees_currency_side())

    # Test sell order
    market_sell = SellMarketOrder(trader_inst)
    sell_order_price = decimal.Decimal(80)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=sell_order_price,
                       quantity=max_buy_quantity * leverage,   # x2 apply leverage (restore leveraged_max_buy_quantity)
                       price=sell_order_price,
                       fees_currency_side=default_contract.get_fees_currency_side())

    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') - \
           max_buy_quantity * buy_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == max_buy_quantity * buy_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO

    # Open long position
    await fill_market_order(market_buy)
    fees = market_buy.get_total_fees(market_buy.market)
    assert fees > constants.ZERO

    # warning here: due to issues on decimal.Decimal precision (22+th digit, no impact), keep real order in operations
    # (the next line will fail if fees are at the end)
    # error if in this order: decimal.Decimal('1000') - max_buy_quantity * buy_order_price - fees:
    #   Decimal('1.1905713144226927686775873') != Decimal('1.190571314422692768677587159')
    # a second rounding issue appears due to leverage computations: re-apply them to end up with the same value
    # error if "leverage / leverage" is not applied :
    #   Decimal('1.1905713144226927686775873') != Decimal('1.1905713144226927686775873')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') - \
           fees - max_buy_quantity * buy_order_price * leverage / leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000') - fees
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == \
           max_buy_quantity * buy_order_price * leverage / leverage

    # change market price to sell order price
    await _update_position_mark_price(exchange_manager_inst, sell_order_price, DEFAULT_FUTURE_SYMBOL)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') - \
           fees - max_buy_quantity * buy_order_price * leverage / leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000') - fees + \
           max_buy_quantity * sell_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == \
           max_buy_quantity * buy_order_price * leverage / leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == \
           max_buy_quantity * (sell_order_price - buy_order_price) * leverage

    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') - \
           fees - max_buy_quantity * buy_order_price * leverage / leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000') - fees + \
           max_buy_quantity * sell_order_price * leverage / leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == \
           max_buy_quantity * buy_order_price * leverage / leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == \
           max_buy_quantity * (sell_order_price - buy_order_price) * leverage

    # Close long position with gains
    await fill_market_order(market_sell)
    sell_fees = market_sell.get_total_fees(market_sell.market)
    assert sell_fees > constants.ZERO

    # warning here: due to issues on decimal.Decimal precision (22+th digit, no impact), keep real order in operations
    # keep order of operations to avoid rounding issues
    fees_result = - fees - sell_fees
    position_results = max_buy_quantity * leverage * (sell_order_price - buy_order_price)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') + \
           fees_result + position_results
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000') + \
           fees_result + position_results
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO


async def test_update_portfolio_data_from_orders_with_max_short_size_linear_contract(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    leverage = decimal.Decimal(2)
    default_contract.set_current_leverage(leverage)

    # Test sell order
    market_sell = SellMarketOrder(trader_inst)
    sell_order_price = decimal.Decimal(40)

    leveraged_max_sell_quantity = get_max_order_quantity_for_price(
        LinearPosition(trader_inst, default_contract),
        portfolio_manager.portfolio.get_currency_portfolio("USDT").available,
        sell_order_price,
        enums.PositionSide.SHORT,
        DEFAULT_FUTURE_SYMBOL
    )
    assert 49 < leveraged_max_sell_quantity < 50   # with a x1 leverage, it would be 24 < max_buy_quantity < 25

    max_sell_quantity = leveraged_max_sell_quantity / leverage
    # warning here: same rounding issues as in test_update_portfolio_data_from_orders_with_max_long_size_linear_contract
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=sell_order_price,
                       quantity=max_sell_quantity * leverage,    # apply leverage (restore leveraged_max_sell_quantity)
                       price=sell_order_price,
                       fees_currency_side=default_contract.get_fees_currency_side())

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    buy_order_price = decimal.Decimal(50)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=buy_order_price,
                      quantity=max_sell_quantity * leverage,    # apply leverage (restore leveraged_max_sell_quantity)
                      price=buy_order_price,
                      fees_currency_side=default_contract.get_fees_currency_side())

    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') - \
           max_sell_quantity * sell_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == \
           max_sell_quantity * sell_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO

    # Open short position
    await fill_market_order(market_sell)
    fees = market_sell.get_total_fees(market_sell.market)
    assert fees > constants.ZERO

    # warning here: same rounding issues as in test_update_portfolio_data_from_orders_with_max_long_size_linear_contract
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') - \
           fees - max_sell_quantity * sell_order_price * leverage / leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000') - fees
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == \
           max_sell_quantity * sell_order_price * leverage / leverage

    # change market price with buy price
    await _update_position_mark_price(exchange_manager_inst, buy_order_price, DEFAULT_FUTURE_SYMBOL)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') - \
           fees - max_sell_quantity * sell_order_price * leverage / leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000') - fees + \
           max_sell_quantity * (sell_order_price - buy_order_price) * leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == \
           max_sell_quantity * sell_order_price * leverage / leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == \
           max_sell_quantity * (sell_order_price - buy_order_price) * leverage

    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') - \
           fees - max_sell_quantity * sell_order_price * leverage / leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000') - fees + \
           max_sell_quantity * (sell_order_price - buy_order_price) * leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == \
           max_sell_quantity * sell_order_price * leverage / leverage
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == \
           max_sell_quantity * (sell_order_price - buy_order_price) * leverage

    # Close short position with gains
    await fill_market_order(market_buy)
    buy_fees = market_buy.get_total_fees(market_buy.market)
    assert buy_fees > constants.ZERO

    # warning here: same rounding issues as in test_update_portfolio_data_from_orders_with_max_long_size_linear_contract
    # keep order of operations to avoid rounding issues
    fees_result = - fees - buy_fees
    position_results = max_sell_quantity * leverage * (sell_order_price - buy_order_price)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') + \
           fees_result + position_results
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000') + \
           fees_result + position_results
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO


async def test_update_portfolio_data_from_orders_with_max_long_size_inverse_contract(
        future_trader_simulator_with_default_inverse):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_inverse
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    leverage = decimal.Decimal(5)
    default_contract.set_current_leverage(leverage)

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    buy_order_price = decimal.Decimal(40)
    leveraged_max_buy_quantity = get_max_order_quantity_for_price(
        InversePosition(trader_inst, default_contract),
        portfolio_manager.portfolio.get_currency_portfolio("BTC").available,
        buy_order_price,
        enums.PositionSide.LONG,
        DEFAULT_FUTURE_SYMBOL
    )
    assert 1990 < leveraged_max_buy_quantity < 2000   # with a x1 leverage, it would be ~ 390 < max_buy_quantity < 400

    max_buy_quantity = leveraged_max_buy_quantity / leverage
    # WARNING:
    # Later on in this test, use max_buy_quantity only to avoid last digit rounding issues on later asserts
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=buy_order_price,
                      quantity=max_buy_quantity * leverage,    # apply leverage (restore leveraged_max_buy_quantity)
                      price=buy_order_price,
                      fees_currency_side=default_contract.get_fees_currency_side())

    # Test sell order
    market_sell = SellMarketOrder(trader_inst)
    sell_order_price = decimal.Decimal(80)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=sell_order_price,
                       quantity=max_buy_quantity * leverage,   # apply leverage (restore leveraged_max_buy_quantity)
                       price=sell_order_price,
                       fees_currency_side=default_contract.get_fees_currency_side())

    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10') - \
           max_buy_quantity / buy_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == max_buy_quantity / buy_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO

    # Open long position
    await fill_market_order(market_buy)
    fees = market_buy.get_total_fees(market_buy.currency)
    assert fees > constants.ZERO

    # warning here: same rounding issues as in test_update_portfolio_data_from_orders_with_max_long_size_linear_contract
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10') - \
           fees - max_buy_quantity * leverage / leverage / buy_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10') - fees
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == \
           max_buy_quantity * leverage / leverage / buy_order_price

    # change market price to sell order price
    await _update_position_mark_price(exchange_manager_inst, sell_order_price, DEFAULT_FUTURE_SYMBOL)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10') - \
           fees - max_buy_quantity * leverage / leverage / buy_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10') - fees + \
           (max_buy_quantity * leverage / sell_order_price)
    # (equivalent to this statement)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10') - fees + \
           5 * max_buy_quantity * (1 / buy_order_price - 1 / sell_order_price)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == \
           max_buy_quantity * leverage / leverage / buy_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == \
           max_buy_quantity * leverage * (1 / buy_order_price - 1 / sell_order_price)

    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10') - \
           fees - max_buy_quantity * leverage / leverage / buy_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10') - fees + \
           (max_buy_quantity * leverage / sell_order_price)
    # (equivalent to this statement)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10') - fees + \
           5 * max_buy_quantity * (1 / buy_order_price - 1 / sell_order_price)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == \
           max_buy_quantity * leverage / leverage / buy_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == \
           max_buy_quantity * leverage * (1 / buy_order_price - 1 / sell_order_price)

    # Close long position with gains
    await fill_market_order(market_sell)
    sell_fees = market_sell.get_total_fees(market_sell.currency)
    assert sell_fees > constants.ZERO

    # warning here: same rounding issues as in test_update_portfolio_data_from_orders_with_max_long_size_linear_contract
    # keep order of operations to avoid rounding issues
    fees_result = - fees - sell_fees
    position_results = max_buy_quantity * leverage * (1 / buy_order_price - 1 / sell_order_price)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10') + \
           fees_result + position_results
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10') + \
           fees_result + position_results
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO


async def test_update_portfolio_data_from_orders_with_max_short_size_inverse_contract(
        future_trader_simulator_with_default_inverse):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_inverse
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    leverage = decimal.Decimal(3)
    default_contract.set_current_leverage(leverage)

    # Test sell order
    market_sell = SellMarketOrder(trader_inst)
    sell_order_price = decimal.Decimal(40)
    leveraged_max_sell_quantity = get_max_order_quantity_for_price(
        InversePosition(trader_inst, default_contract),
        portfolio_manager.portfolio.get_currency_portfolio("BTC").available,
        sell_order_price,
        enums.PositionSide.SHORT,
        DEFAULT_FUTURE_SYMBOL
    )
    assert 1195 < leveraged_max_sell_quantity < 1200   # with a x1 leverage, it would be 390 < max_buy_quantity < 400

    max_sell_quantity = leveraged_max_sell_quantity / leverage
    # WARNING:
    # Later on in this test, use max_buy_quantity only to avoid last digit rounding issues on later asserts
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=sell_order_price,
                       quantity=max_sell_quantity * leverage,    # apply leverage (restore leveraged_max_sell_quantity)
                       price=sell_order_price,
                       fees_currency_side=default_contract.get_fees_currency_side())

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    buy_order_price = decimal.Decimal(50)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=buy_order_price,
                      quantity=max_sell_quantity * leverage,    # apply leverage (restore leveraged_max_sell_quantity)
                      price=buy_order_price,
                      fees_currency_side=default_contract.get_fees_currency_side())

    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10') - \
           max_sell_quantity / sell_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == \
           max_sell_quantity / sell_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO

    # Open short position
    await fill_market_order(market_sell)
    fees = market_sell.get_total_fees(market_sell.currency)
    assert fees > constants.ZERO

    # warning here: same rounding issues as in test_update_portfolio_data_from_orders_with_max_long_size_linear_contract
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10') - \
           fees - (max_sell_quantity * leverage / leverage / sell_order_price)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10') - fees
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == \
           max_sell_quantity * leverage / leverage / sell_order_price

    # change market price with buy price
    await _update_position_mark_price(exchange_manager_inst, buy_order_price, DEFAULT_FUTURE_SYMBOL)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10') - \
           fees - max_sell_quantity * leverage / leverage / sell_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10') - fees + \
           max_sell_quantity * leverage * (1 / buy_order_price - 1 / sell_order_price)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == \
           max_sell_quantity * leverage / leverage / sell_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == \
           max_sell_quantity * leverage * (1 / buy_order_price - 1 / sell_order_price)

    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10') - \
           fees - max_sell_quantity * leverage / leverage / sell_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10') - fees + \
           max_sell_quantity * leverage * (1 / buy_order_price - 1 / sell_order_price)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == \
           max_sell_quantity * leverage / leverage / sell_order_price
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == \
           max_sell_quantity * leverage * (1 / buy_order_price - 1 / sell_order_price)

    # Close short position with gains
    await fill_market_order(market_buy)
    buy_fees = market_buy.get_total_fees(market_buy.currency)
    assert buy_fees > constants.ZERO

    # warning here: same rounding issues as in test_update_portfolio_data_from_orders_with_max_long_size_linear_contract
    # keep order of operations to avoid rounding issues
    fees_result = - buy_fees - fees
    position_results = max_sell_quantity * leverage * ( 1 / buy_order_price - 1 / sell_order_price)

    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10') + \
           fees_result + position_results
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10') + \
           fees_result + position_results
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == constants.ZERO


async def test_update_portfolio_data_from_order_with_huge_loss_on_filled_orders_linear_contract(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(2))

    # Test sell order
    market_sell = SellMarketOrder(trader_inst)
    sell_order_price = decimal.Decimal(10)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=sell_order_price,
                       quantity=decimal.Decimal(25),
                       price=sell_order_price)

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    buy_order_price = decimal.Decimal(13)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=buy_order_price,
                      quantity=decimal.Decimal(10),
                      price=buy_order_price)

    closing_market_buy = BuyMarketOrder(trader_inst)
    closing_market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                              symbol=DEFAULT_FUTURE_SYMBOL,
                              current_price=buy_order_price,
                              quantity=decimal.Decimal(15),
                              price=buy_order_price)

    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('875.0')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == decimal.Decimal('125')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO

    # Open short position
    await fill_market_order(market_sell)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('874.9')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('999.9')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal('125')

    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('874.9')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('999.9')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal('125')

    # Close short position with loss
    await fill_market_order(market_buy)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('894.848')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('924.848')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal('75')

    # Close short position with loss
    await fill_market_order(closing_market_buy)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('924.77')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('924.77')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO


async def test_update_portfolio_from_liquidated_position_with_orders_on_short_position_linear_contract(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(50))
    trader_inst.exchange_manager.exchange.set_pair_future_contract(DEFAULT_FUTURE_SYMBOL, default_contract)

    position_inst = LinearPosition(trader_inst, default_contract)
    await position_inst.initialize()
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    await position_inst.update(update_size=decimal.Decimal(-100), mark_price=decimal.Decimal(99))
    exchange_manager_inst.exchange_personal_data.positions_manager.upsert_position_instance(position_inst)

    # Test sell order
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(99),
                       quantity=decimal.Decimal(5),
                       price=decimal.Decimal(99))

    # test sell order creation
    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('792.1')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == decimal.Decimal('9.9')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal('198')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO

    await fill_market_order(market_sell)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('791.902')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('999.802')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')

    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(90),
                      quantity=decimal.Decimal(10),
                      price=decimal.Decimal(90))

    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('791.902')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal('207.9')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('999.802')

    await fill_market_order(market_buy)  # reducing position with positive PNL
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('901.342')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1944.442')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal('188.1')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == decimal.Decimal(str(855))

    portfolio_manager.portfolio.get_currency_portfolio("BTC").wallet_balance = decimal.Decimal('10')
    portfolio_manager.portfolio.get_currency_portfolio("USDT").wallet_balance = decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal(str(188.1))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == decimal.Decimal(str(855.0))

    await position_inst.update(mark_price=position_inst.liquidation_price)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('811.9')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('811.9')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')


async def test_update_portfolio_from_funding_with_long_position(
        future_trader_simulator_with_default_inverse):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_inverse
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(10))

    position_inst = InversePosition(trader_inst, default_contract)
    await position_inst.initialize()
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    await position_inst.update(update_size=decimal.Decimal(10), mark_price=decimal.Decimal(100))
    exchange_manager_inst.exchange_personal_data.positions_manager.upsert_position_instance(position_inst)

    # long position holders have to pay the short position holders
    portfolio_manager.portfolio.update_portfolio_from_funding(position=position_inst,
                                                              funding_rate=decimal.Decimal(0.0001))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '9.989989999999999999999520783')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '9.999989999999999999999520783')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == decimal.Decimal('0.01')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO

    position_inst = InversePosition(trader_inst, default_contract)
    await position_inst.initialize()
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    await position_inst.update(update_size=decimal.Decimal(10), mark_price=decimal.Decimal(100))
    exchange_manager_inst.exchange_personal_data.positions_manager.upsert_position_instance(position_inst)

    # short position holders have to pay the long position holders
    portfolio_manager.portfolio.update_portfolio_from_funding(position=position_inst,
                                                              funding_rate=decimal.Decimal(-0.0002))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '9.98001000000000000000047922')  # = previous available + funding - position increased size
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '10.00001000000000000000047922')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == decimal.Decimal('0.02')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO

    # should raise PortfolioNegativeValueError when funding fee > available
    with pytest.raises(errors.PortfolioNegativeValueError):
        portfolio_manager.portfolio.update_portfolio_from_funding(position=position_inst,
                                                                  funding_rate=decimal.Decimal(1000))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '9.98001000000000000000047922')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '10.00001000000000000000047922')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == decimal.Decimal('0.02')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO


async def test_update_portfolio_from_funding_with_short_position(
        future_trader_simulator_with_default_inverse):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_inverse
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(10))

    position_inst = InversePosition(trader_inst, default_contract)
    await position_inst.initialize()
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    await position_inst.update(update_size=decimal.Decimal(-10), mark_price=decimal.Decimal(100))
    exchange_manager_inst.exchange_personal_data.positions_manager.upsert_position_instance(position_inst)

    # long position holders have to pay the short position holders
    portfolio_manager.portfolio.update_portfolio_from_funding(position=position_inst,
                                                              funding_rate=decimal.Decimal(0.0003))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '9.99002999999999999999737189')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '10.00002999999999999999737189')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == decimal.Decimal(str(0.01))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO

    position_inst = InversePosition(trader_inst, default_contract)
    await position_inst.initialize()
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    await position_inst.update(update_size=decimal.Decimal(-10), mark_price=decimal.Decimal(100))
    exchange_manager_inst.exchange_personal_data.positions_manager.upsert_position_instance(position_inst)

    # short position holders have to pay the long position holders
    portfolio_manager.portfolio.update_portfolio_from_funding(position=position_inst,
                                                              funding_rate=decimal.Decimal(-0.0004))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '9.979989999999999999995455021')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '9.999989999999999999995455021')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == decimal.Decimal('0.02')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO

    # should raise PortfolioNegativeValueError when funding fee > available
    with pytest.raises(errors.PortfolioNegativeValueError):
        portfolio_manager.portfolio.update_portfolio_from_funding(position=position_inst,
                                                                  funding_rate=decimal.Decimal(-2378))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(
        '9.979989999999999999995455021')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(
        '9.999989999999999999995455021')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").position_margin == decimal.Decimal('0.02')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").unrealized_pnl == constants.ZERO


async def test_update_portfolio_data_with_fees(future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(10))

    # Test sell order
    market_sell = SellMarketOrder(trader_inst)
    sell_order_price = decimal.Decimal(10)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=sell_order_price,
                       quantity=decimal.Decimal(25),
                       price=sell_order_price)

    if not os.getenv('CYTHON_IGNORE'):
        with mock.patch.object(market_sell, "get_total_fees", mock.Mock(return_value=5)) as get_total_fees_mock:
            portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('975.0')
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == decimal.Decimal('25')
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO

            # fill order with fees
            get_total_fees_mock.assert_not_called()
            await fill_market_order(market_sell)
            assert get_total_fees_mock.call_count == 2  # once at fill, once to update position size
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('970.0')
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('995')
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal('25')
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO


async def test_update_portfolio_data_with_fees_long_position(future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    leverage = decimal.Decimal(2)
    default_contract.set_current_leverage(leverage)

    # data from real test on Bybit from 02.02.2020 using 2 market orders:
    # entry price: 38497.5, exit price: 38503, quantity: 0.01, leverage: 2x, taker fees: 0.075%
    # profits from position = +0.055, fees = -0.5775, displayed closed pnl = -0.5225
    # Test buy order
    market_buy = BuyMarketOrder(trader_inst)
    market_sell = SellMarketOrder(trader_inst)
    buy_order_price = decimal.Decimal('38497.5')
    sell_order_price = decimal.Decimal('38503')
    quantity = decimal.Decimal('0.01')
    fees_per_order = decimal.Decimal('0.2887')  # 0.2887 = 0.5775 / 2
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=buy_order_price,
                      quantity=quantity,
                      price=buy_order_price)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=sell_order_price,
                       quantity=quantity,
                       price=sell_order_price)

    if not os.getenv('CYTHON_IGNORE'):
        with mock.patch.object(market_buy, "get_total_fees", mock.Mock(return_value=fees_per_order)) \
                as get_total_fees_mock:

            # create buy order
            portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
            position = exchange_manager_inst.exchange_personal_data.positions_manager.get_symbol_positions()[0]
            required_funds = buy_order_price * quantity / leverage
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') \
                   - required_funds
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == required_funds
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO
            assert position.size == constants.ZERO

            # fill order with fees
            get_total_fees_mock.assert_not_called()
            await fill_market_order(market_buy)
            assert get_total_fees_mock.call_count == 2  # once at fill, once to update position size
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') \
                   - required_funds - fees_per_order
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000') \
                   - fees_per_order
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == \
                   required_funds
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO
            assert position.size == quantity
            assert position.realised_pnl == -fees_per_order

        with mock.patch.object(market_sell, "get_total_fees", mock.Mock(return_value=fees_per_order)) \
                as get_total_fees_mock:

            # create sell order
            portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
            position = exchange_manager_inst.exchange_personal_data.positions_manager.get_symbol_positions()[0]
            # reducing position: no change in available
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') \
                   - required_funds - fees_per_order
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000') \
                   - fees_per_order
            # reducing position: no change in order margin
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == \
                   required_funds
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO
            assert position.size == quantity
            assert position.realised_pnl == -fees_per_order

            # fill order with fees
            profits = (sell_order_price - buy_order_price) * quantity
            total_pnl = - fees_per_order - fees_per_order + profits   # -0.5225
            get_total_fees_mock.assert_not_called()
            await fill_market_order(market_sell)
            assert get_total_fees_mock.call_count == 2  # once at fill, once to update position size
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000') \
                   + total_pnl
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000') \
                   + total_pnl
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == constants.ZERO
            assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO
            assert position.size == constants.ZERO
            assert position.realised_pnl == total_pnl


async def test_update_portfolio_reduce_size_with_market_sell_long_linear_contract(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(10))

    position_inst = LinearPosition(trader_inst, default_contract)
    await position_inst.initialize()
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    await position_inst.update(update_size=decimal.Decimal(10), mark_price=decimal.Decimal(101))
    exchange_manager_inst.exchange_personal_data.positions_manager.upsert_position_instance(position_inst)

    # Test sell order
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(101),
                       quantity=decimal.Decimal(25),
                       price=decimal.Decimal(101))

    # test sell order creation
    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(899))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(str(10))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(str(10))

    # fill order
    await fill_market_order(market_sell)
    fees = market_sell.get_total_fees(market_sell.market)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal("1000") - fees
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str("848.5")) - fees
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal('151.5')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(str(10))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(str(10))


async def test_update_portfolio_reduce_size_with_market_buy_short_linear_contract(
        future_trader_simulator_with_default_linear):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(10))

    position_inst = LinearPosition(trader_inst, default_contract)
    await position_inst.initialize()
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    await position_inst.update(update_size=constants.ZERO, mark_price=decimal.Decimal(10))
    exchange_manager_inst.exchange_personal_data.positions_manager.upsert_position_instance(position_inst)

    # initialize short
    market_sell = SellMarketOrder(trader_inst)
    market_sell.update(order_type=enums.TraderOrderType.SELL_MARKET,
                       symbol=DEFAULT_FUTURE_SYMBOL,
                       current_price=decimal.Decimal(10),
                       quantity=decimal.Decimal(70),
                       price=decimal.Decimal(10))
    portfolio_manager.portfolio.update_portfolio_available(market_sell, True)
    await fill_market_order(market_sell)

    # Test reducing buy order
    market_buy = BuyMarketOrder(trader_inst)
    market_buy.update(order_type=enums.TraderOrderType.BUY_MARKET,
                      symbol=DEFAULT_FUTURE_SYMBOL,
                      current_price=decimal.Decimal(10),
                      quantity=decimal.Decimal(20),
                      price=decimal.Decimal(10))

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    # 1000 - 70 - fees
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('929.72')
    # 1000 - fees
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('999.72')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal('70')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(str(10))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(str(10))

    # fill order
    await fill_market_order(market_buy)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('999.64')
    # 1000 - 70 - fees + 20
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('949.64')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").order_margin == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").position_margin == decimal.Decimal('50')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").unrealized_pnl == constants.ZERO
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal(str(10))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(str(10))


async def test_update_portfolio_from_pnl_with_long_inverse_contract(future_trader_simulator_with_default_inverse):
    config, exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_inverse
    portfolio_manager = exchange_manager_inst.exchange_personal_data.portfolio_manager
    default_contract.set_current_leverage(decimal.Decimal(10))

    position_inst = InversePosition(trader_inst, default_contract)
    await position_inst.initialize()
    position_inst.update_from_raw({enums.ExchangeConstantsPositionColumns.SYMBOL.value: DEFAULT_FUTURE_SYMBOL})
    await position_inst.update(update_size=decimal.Decimal(-70), mark_price=decimal.Decimal(10))
    exchange_manager_inst.exchange_personal_data.positions_manager.upsert_position_instance(position_inst)

    position_inst.unrealized_pnl = decimal.Decimal("0.1")
    portfolio_manager.portfolio.update_portfolio_from_pnl(position_inst)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(1000))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('9.3')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10.1')

    position_inst.unrealized_pnl = decimal.Decimal("-0.3")
    portfolio_manager.portfolio.update_portfolio_from_pnl(position_inst)
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal(str(1000))
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal(str(1000))
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('9.3')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal(str('9.7'))


def _disable_fees(exchange_manager):
    exchange_manager.exchange.connector.client.calculate_fee = mock.Mock(return_value={
        enums.FeePropertyColumns.TYPE.value: "maker",
        enums.FeePropertyColumns.CURRENCY.value: "USDT",
        enums.FeePropertyColumns.RATE.value: 0,
        enums.FeePropertyColumns.COST.value: constants.ZERO,
    })



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport async_channel.consumer as consumer
cimport octobot_backtesting.importers as importers
cimport octobot_trading.exchange_data.ohlcv.channel.ohlcv_updater as ohlcv_updater


cdef class OHLCVUpdaterSimulator(ohlcv_updater.OHLCVUpdater):
    cdef importers.ExchangeDataImporter exchange_data_importer

    cdef str exchange_name

    cdef double initial_timestamp
    cdef double last_timestamp_pushed
    cdef dict time_frames_to_second

    cdef public consumer.Consumer time_consumer

    cdef object future_candle_time_frame
    cdef int future_candle_sec_length

    cdef dict last_candles_by_pair_by_time_frame
    cdef bint require_last_init_candles_pairs_push
    cdef list traded_pairs
    cdef list traded_time_frame



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import collections

import octobot_commons.logging as logging

import octobot_trading.util as util


class RecentTradesManager(util.Initializable):
    MAX_RECENT_TRADES_COUNT = 100
    MAX_LIQUIDATIONS_COUNT = 20

    def __init__(self):
        super().__init__()
        self.logger = logging.get_logger(self.__class__.__name__)
        self.recent_trades = collections.deque(maxlen=self.MAX_RECENT_TRADES_COUNT)
        self.liquidations = collections.deque(maxlen=self.MAX_LIQUIDATIONS_COUNT)
        self._reset_recent_trades()

    async def initialize_impl(self):
        self._reset_recent_trades()

    def set_all_recent_trades(self, recent_trades):
        if recent_trades:
            self.recent_trades = recent_trades
            return self.recent_trades

    def add_new_trades(self, recent_trades):
        if recent_trades:
            new_recent_trades: list = [
                trade
                for trade in recent_trades
                if trade not in self.recent_trades]
            self.recent_trades.extend(new_recent_trades)
            return new_recent_trades

    def add_new_liquidations(self, liquidations):
        if liquidations:
            new_liquidations: list = [
                liquidation
                for liquidation in liquidations
                if liquidation not in self.liquidations]
            self.liquidations.extend(new_liquidations)
            return new_liquidations

    def _reset_recent_trades(self):
        self.recent_trades = collections.deque(maxlen=self.MAX_RECENT_TRADES_COUNT)
        self.liquidations = collections.deque(maxlen=self.MAX_LIQUIDATIONS_COUNT)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cpdef object get_margin_exchange_class(str exchange_name, object tentacles_setup_config)
cpdef object get_future_exchange_class(str exchange_name, object tentacles_setup_config)
cpdef object get_spot_exchange_class(str exchange_name, object tentacles_setup_config)
cpdef str get_order_side(object order_type)
cpdef void log_time_sync_error(object logger, str exchange_name, object error, str details)
cpdef str get_partners_explanation_message()
cpdef object get_exchange_type(object exchange_manager_instance)
cpdef str get_default_exchange_type(str exchange_name)
cpdef list get_supported_exchange_types(str exchange_name)
cpdef object get_exchange_class_from_name(object exchange_parent_class,
                                          str exchange_name,
                                          object tentacles_setup_config,
                                          bint enable_default,
                                          bint strict_name_matching=*)

cdef object search_exchange_class_from_exchange_name(object exchange_class,
                                                     str exchange_name,
                                                     object tentacles_setup_config,
                                                     bint enable_default=*)
cdef bint _is_exchange_candidate_matching(object exchange_candidate,
                                          str exchange_name,
                                          object tentacles_setup_config,
                                          bint enable_default=*)
cdef str _get_docs_url()
cdef str _get_exchanges_docs_url()
cdef str _get_time_sync_error_message(str exchange_name, str details)
cdef dict _get_minimal_exchange_config(str exchange_name, dict exchange_config)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cpdef bint is_ms_valid(object value, bint zero_valid=*)
cdef bint check_market_status_limits(market_limit)
cdef bint check_market_status_values(market_limit, bint zero_valid=*)
cdef object get_markets_limit(dict market_limit)
cdef void calculate_amounts(dict market_limit)
cdef void calculate_costs(dict market_limit)
cdef void calculate_prices(dict market_limit)
cdef void fix_market_status_limits_from_current_data(dict market_limit)

cdef class ExchangeMarketStatusFixer:
    cdef public dict market_status
    cdef object logger
    cdef object price_example
    cdef dict market_status_specific

    cdef void _fix_market_status_precision(self)
    cdef void _fix_market_status_limits(self)
    cdef tuple _calculate_amount(self)
    cdef void _fix_market_status_limits_with_price(self)
    cdef void _fix_market_status_precision_with_price(self)
    cdef double _get_price_precision(self)
    cdef void _fix_market_status_limits_with_specific(self)
    cdef void _fix_market_status_precision_with_specific(self)
    cdef void _fix_typing(self)
    cdef void _convert_values_to_float(self, dict element, list parent_keys, list key_whitelist)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock
import os
import sortedcontainers
import decimal

import octobot_commons.enums as commons_enums
import octobot_trading.personal_data as personal_data
import octobot_trading.constants as constants

from tests.exchanges import backtesting_trader_with_historical_pf_value_manager, \
    backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def historical_portfolio_value_manager(backtesting_trader_with_historical_pf_value_manager):
    config, exchange_manager, trader = backtesting_trader_with_historical_pf_value_manager
    return exchange_manager.exchange_personal_data.portfolio_manager.historical_portfolio_value_manager


def test_constructor(historical_portfolio_value_manager):
    assert historical_portfolio_value_manager is not None
    assert historical_portfolio_value_manager.portfolio_manager is not None
    assert historical_portfolio_value_manager.saved_time_frames == constants.DEFAULT_SAVED_HISTORICAL_TIMEFRAMES
    assert historical_portfolio_value_manager.historical_portfolio_value == sortedcontainers.SortedDict()
    assert historical_portfolio_value_manager.starting_time is not None
    assert historical_portfolio_value_manager.starting_time == historical_portfolio_value_manager.last_update_time
    assert historical_portfolio_value_manager.starting_portfolio is None
    assert historical_portfolio_value_manager.ending_portfolio is None


async def test_initialize(historical_portfolio_value_manager):
    # run_dbs_identifier is None: does nothing
    await historical_portfolio_value_manager.initialize()
    assert historical_portfolio_value_manager.historical_portfolio_value == sortedcontainers.SortedDict()

    if os.getenv('CYTHON_IGNORE'):
        return
    with mock.patch.object(historical_portfolio_value_manager, "_reload_historical_portfolio_value", mock.AsyncMock()) \
        as _reload_historical_portfolio_value_mock:
        historical_portfolio_value_manager.is_initialized = False
        await historical_portfolio_value_manager.initialize()
        _reload_historical_portfolio_value_mock.assert_called_once()


async def test_on_new_value(historical_portfolio_value_manager):
    timestamp = 1648462965  # Monday 28 March 2022 10:22:45 UTC
    exchange_time = timestamp + 10
    day_timestamp = 1648425600  # Monday 28 March 2022 00:00:00 UTC
    # force exchange current time
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = exchange_time
    value_by_currency = {
        "BTC": 1,
        "USD": 3000
    }

    # new timestamp, too late in the day, ignore it
    too_late_timestamp = 1648504800  # Monday 28 March 2022 22:00:00 UTC
    assert await historical_portfolio_value_manager.on_new_value(too_late_timestamp, value_by_currency,
                                                                 save_changes=False) is False
    assert historical_portfolio_value_manager.historical_portfolio_value == sortedcontainers.SortedDict()

    # new timestamp early enough in the day, add it
    assert await historical_portfolio_value_manager.on_new_value(timestamp, value_by_currency,
                                                                 save_changes=False) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [day_timestamp]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            day_timestamp, value_by_currency)

    # new currency value, same time
    value_by_currency_2 = {
        "BTC": 1,
        "USD": 3000,
        "HELLO": 666
    }
    assert await historical_portfolio_value_manager.on_new_value(timestamp, value_by_currency_2,
                                                                 save_changes=False) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [day_timestamp]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            day_timestamp, value_by_currency_2)

    # change currency value, same time
    value_by_currency_3 = {
        "BTC": 1.5,
        "USD": 3001,
        "HELLO": 666
    }
    # no force update: no change
    assert await historical_portfolio_value_manager.on_new_value(timestamp, value_by_currency_3,
                                                                 save_changes=False) is False
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [day_timestamp]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            day_timestamp, value_by_currency_2)
    # force update: actual change
    assert await historical_portfolio_value_manager.on_new_value(timestamp, value_by_currency_3,
                                                                 save_changes=False, force_update=True) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [day_timestamp]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            day_timestamp, value_by_currency_3)

    # new timestamp, same day: ignore it
    timestamp_same_day = timestamp + 1000
    assert await historical_portfolio_value_manager.on_new_value(timestamp_same_day, value_by_currency_3,
                                                                 save_changes=False) is False
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [day_timestamp]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            day_timestamp, value_by_currency_3)

    # timestamp on new day, same day: ignore
    timestamp_new_day = 1648540800  # Tuesday 29 March 2022 08:00:00
    assert await historical_portfolio_value_manager.on_new_value(timestamp_new_day, value_by_currency_3,
                                                                 save_changes=False) is False
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [day_timestamp]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            day_timestamp, value_by_currency_3)

    # force exchange current time
    new_day_timestamp = 1648512000
    new_day_exchange_time = new_day_timestamp + 3001
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = new_day_exchange_time
    # timestamp on new day, new day: add it
    assert await historical_portfolio_value_manager.on_new_value(timestamp_new_day, value_by_currency_3,
                                                                 save_changes=False) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [day_timestamp,
                                                                                          new_day_timestamp]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            day_timestamp, value_by_currency_3)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[1],
                            new_day_timestamp, value_by_currency_3)

    value_by_currency_4 = {
        "BTC": 9,
        "USD": 1111,
        "HELLO": 666
    }
    # force update: actual change
    assert await historical_portfolio_value_manager.on_new_value(timestamp_new_day, value_by_currency_4,
                                                                 save_changes=False, force_update=True) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [day_timestamp,
                                                                                          new_day_timestamp]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            day_timestamp, value_by_currency_3)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[1],
                            new_day_timestamp, value_by_currency_4)


async def test_on_new_value_with_max_size_reached(historical_portfolio_value_manager):
    timestamp = 1648462965  # Monday 28 March 2022 10:22:45 UTC
    exchange_time = timestamp + 10
    day_timestamp = 1648425600  # Monday 28 March 2022 00:00:00 UTC
    # force exchange current time
    historical_portfolio_value_manager.max_history_size = 2
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = exchange_time
    value_by_currency = {
        "BTC": decimal.Decimal(1),
        "USD": decimal.Decimal(3000)
    }
    # new timestamp early enough in the day, add it
    assert await historical_portfolio_value_manager.on_new_value(timestamp, value_by_currency,
                                                                 save_changes=False) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [day_timestamp]

    new_day_timestamp = 1648512000
    timestamp_new_day = new_day_timestamp + 1
    value_by_currency_2 = {
        "BTC": decimal.Decimal(11),
        "USD": decimal.Decimal(30003)
    }
    new_day_exchange_time = new_day_timestamp + 3001
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = new_day_exchange_time
    # timestamp on new day, new day: add it
    assert await historical_portfolio_value_manager.on_new_value(timestamp_new_day, value_by_currency_2,
                                                                 save_changes=False) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [day_timestamp,
                                                                                          new_day_timestamp]
    float_value_by_currency = {
        "BTC": 1,
        "USD": 3000
    }
    float_value_by_currency_2 = {
        "BTC": 11,
        "USD": 30003
    }
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            day_timestamp, float_value_by_currency)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[1],
                            new_day_timestamp, float_value_by_currency_2)

    other_new_day_timestamp = 1648598400  # Wednesday 30 March 2022 00:00:00 UTC
    other_timestamp_new_day = other_new_day_timestamp + 1
    value_by_currency_3 = {
        "BTC": decimal.Decimal("111.111"),
        "USD": decimal.Decimal("300033.111")
    }
    float_value_by_currency_3 = {
        "BTC": 111.111,
        "USD": 300033.111
    }
    new_day_exchange_time = other_new_day_timestamp + 3001
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = new_day_exchange_time
    # timestamp on new day, new day: add it but remove the oldest one as max size is reached
    assert await historical_portfolio_value_manager.on_new_value(other_timestamp_new_day, value_by_currency_3,
                                                                 save_changes=False) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [new_day_timestamp,
                                                                                          other_new_day_timestamp]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            new_day_timestamp, float_value_by_currency_2)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[1],
                            other_new_day_timestamp, float_value_by_currency_3)


async def test_on_new_value_multiple_time_frames(historical_portfolio_value_manager):
    historical_portfolio_value_manager.saved_time_frames = [
        commons_enums.TimeFrames.ONE_DAY, commons_enums.TimeFrames.ONE_HOUR
    ]
    timestamp = 1648462965  # Monday 28 March 2022 10:22:45 UTC
    day_timestamp = 1648425600  # Monday 28 March 2022 00:00:00 UTC
    hour_timestamp = 1648461600  # Monday 28 March 2022 00:10:00 UTC
    exchange_time = timestamp + 10
    # force exchange current time
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = exchange_time
    value_by_currency = {
        "BTC": 1,
        "USD": 3000
    }

    too_late_timestamp = 1648504800  # Monday 28 March 2022 22:00:00 UTC
    # new timestamp, too late in the day, ignore it
    assert await historical_portfolio_value_manager.on_new_value(too_late_timestamp, value_by_currency,
                                                                 save_changes=False) is False
    assert historical_portfolio_value_manager.historical_portfolio_value == sortedcontainers.SortedDict()

    # new timestamp, too late in the day, ignore it BUT not too late for one hour time_frame
    late_hour_timestamp = 1648483200  # Monday 28 March 2022 16:00:00 UTC
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = late_hour_timestamp
    assert await historical_portfolio_value_manager.on_new_value(late_hour_timestamp + 1, value_by_currency,
                                                                 save_changes=False) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [late_hour_timestamp]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            late_hour_timestamp, value_by_currency)

    value_by_currency_2 = {
        "BTC": 10
    }
    # new timestamp early enough in the day, add it
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = hour_timestamp + 3
    assert await historical_portfolio_value_manager.on_new_value(hour_timestamp + 1, value_by_currency_2,
                                                                 save_changes=False) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [
        day_timestamp, hour_timestamp, late_hour_timestamp
    ]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            day_timestamp, value_by_currency_2)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[1],
                            hour_timestamp, value_by_currency_2)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[2],
                            late_hour_timestamp, value_by_currency)

    value_by_currency_3 = {
        "ETH": 55
    }
    # new timestamp that happens to be associated to the same time in 1h and 1d
    early_day_hour_timestamp = 1648512900  # Tuesday 29 March 2022 00:15:00 UTC
    tuesday_timestamp = 1648512000  # Tuesday 29 March 2022 00:00:00 UTC
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = early_day_hour_timestamp
    assert await historical_portfolio_value_manager.on_new_value(early_day_hour_timestamp + 1, value_by_currency_3,
                                                                 save_changes=False) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [
        day_timestamp, hour_timestamp, late_hour_timestamp, tuesday_timestamp
    ]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            day_timestamp, value_by_currency_2)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[1],
                            hour_timestamp, value_by_currency_2)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[2],
                            late_hour_timestamp, value_by_currency)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[3],
                            tuesday_timestamp, value_by_currency_3)


async def test_on_new_values(historical_portfolio_value_manager):
    timestamp = 1648462965  # Monday 28 March 2022 10:22:45 UTC
    exchange_time = timestamp + 10
    friday_timestamp = 1648166400  # Friday 25 March 2022 00:00:00
    saturday_timestamp = 1648252800  # Saturday 26 March 2022 00:00:00
    sunday_timestamp = 1648339200  # Sunday 27 March 2022 00:00:00
    today_timestamp = 1648425600  # Monday 28 March 2022 00:00:00 UTC
    # force exchange current time
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = exchange_time
    friday_value_by_currency = {
        "BTC": 1.1,
        "USD": 3010
    }
    saturday_value_by_currency = {
        "BTC": 1.3
    }
    sunday_value_by_currency = {
        "USD": 3
    }
    today_value_by_currency = {
        "BTC": 11,
        "USD": 54,
        "ETH": 33
    }
    # new timestamp, too late in the day, ignore it
    too_late_timestamp = 1648504800  # Monday 28 March 2022 22:00:00 UTC
    value_by_currency_by_timestamp = {
        friday_timestamp: friday_value_by_currency,     # use exact timestamps for some past timestamps
        sunday_timestamp: sunday_value_by_currency,     # on purpose out of order, value manager is supposed to sort it
        saturday_timestamp + 10: saturday_value_by_currency,
        too_late_timestamp: today_value_by_currency,    # use too_late_timestamp for today timestamp
    }
    # saves historical timestamps but not today's one as it is too late
    assert await historical_portfolio_value_manager.on_new_values(value_by_currency_by_timestamp,
                                                                  save_changes=False) is True
    # also ensure sorting
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [
        friday_timestamp, saturday_timestamp, sunday_timestamp
    ]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            friday_timestamp, friday_value_by_currency)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[1],
                            saturday_timestamp, saturday_value_by_currency)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[2],
                            sunday_timestamp, sunday_value_by_currency)

    # now provide an ok timestamp for today's values
    value_by_currency_by_timestamp = {
        friday_timestamp: friday_value_by_currency,     # use exact timestamps for some past timestamps
        sunday_timestamp: sunday_value_by_currency,     # on purpose out of order, value manager is supposed to sort it
        saturday_timestamp + 10: saturday_value_by_currency,
        timestamp: today_value_by_currency,
    }
    assert await historical_portfolio_value_manager.on_new_values(value_by_currency_by_timestamp,
                                                                  save_changes=False) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [
        friday_timestamp, saturday_timestamp, sunday_timestamp, today_timestamp
    ]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[0],
                            friday_timestamp, friday_value_by_currency)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[1],
                            saturday_timestamp, saturday_value_by_currency)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[2],
                            sunday_timestamp, sunday_value_by_currency)
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[3],
                            today_timestamp, today_value_by_currency)

    # same call, no new timestamp value
    assert await historical_portfolio_value_manager.on_new_values(value_by_currency_by_timestamp,
                                                                  save_changes=False) is False
    sunday_value_by_currency["USD"] = 999
    assert await historical_portfolio_value_manager.on_new_values(value_by_currency_by_timestamp,
                                                                  save_changes=False) is False
    assert await historical_portfolio_value_manager.on_new_values(value_by_currency_by_timestamp,
                                                                  save_changes=False, force_update=True) is True
    assert list(historical_portfolio_value_manager.historical_portfolio_value.keys()) == [
        friday_timestamp, saturday_timestamp, sunday_timestamp, today_timestamp
    ]
    _check_historical_value(list(historical_portfolio_value_manager.historical_portfolio_value.values())[2],
                            sunday_timestamp, sunday_value_by_currency)


async def test_get_historical_value(historical_portfolio_value_manager):
    timestamp = 1648462965  # Monday 28 March 2022 10:22:45 UTC
    exchange_time = timestamp + 10
    friday_timestamp = 1648166400  # Friday 25 March 2022 00:00:00
    saturday_timestamp = 1648252800  # Saturday 26 March 2022 00:00:00
    sunday_timestamp = 1648339200  # Sunday 27 March 2022 00:00:00
    today_timestamp = 1648425600  # Monday 28 March 2022 00:00:00 UTC
    # force exchange current time
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = exchange_time
    friday_value_by_currency = {
        "BTC": 1.1,
        "USD": 3010
    }
    saturday_value_by_currency = {
        "BTC": 1.3
    }
    sunday_value_by_currency = {
        "USD": 3
    }
    today_value_by_currency = {
        "BTC": 11,
        "USD": 54,
        "ETH": 33
    }
    value_by_currency_by_timestamp = {
        friday_timestamp: friday_value_by_currency,     # use exact timestamps for some past timestamps
        sunday_timestamp: sunday_value_by_currency,     # on purpose out of order, value manager is supposed to sort it
        saturday_timestamp + 10: saturday_value_by_currency,
        timestamp: today_value_by_currency,
    }
    await historical_portfolio_value_manager.on_new_values(value_by_currency_by_timestamp, save_changes=False)
    historical_value = historical_portfolio_value_manager.get_historical_value(saturday_timestamp)
    assert historical_value.to_dict() == {
        historical_value.TIMESTAMP_KEY: saturday_timestamp,
        historical_value.VALUES_KEY: saturday_value_by_currency,
    }
    historical_value = historical_portfolio_value_manager.get_historical_value(today_timestamp)
    assert historical_value.to_dict() == {
        historical_value.TIMESTAMP_KEY: today_timestamp,
        historical_value.VALUES_KEY: today_value_by_currency,
    }
    with pytest.raises(KeyError):
        historical_portfolio_value_manager.get_historical_value(1)


async def test_get_historical_values(historical_portfolio_value_manager):
    timestamp = 1648462965  # Monday 28 March 2022 10:22:45 UTC
    exchange_time = timestamp + 10
    friday_timestamp = 1648166400  # Friday 25 March 2022 00:00:00
    saturday_timestamp = 1648252800  # Saturday 26 March 2022 00:00:00
    sunday_timestamp = 1648339200  # Sunday 27 March 2022 00:00:00
    today_timestamp = 1648425600  # Monday 28 March 2022 00:00:00 UTC
    today_hour_timestamp = 1648461600  # Monday 28 March 2022 10:00:00 UTC
    # force exchange current time
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = exchange_time
    historical_portfolio_value_manager.saved_time_frames = [
        commons_enums.TimeFrames.ONE_DAY, commons_enums.TimeFrames.ONE_HOUR
    ]
    friday_value_by_currency = {
        "BTC": 1.1,
        "USD": 3010
    }
    saturday_value_by_currency = {
        "BTC": 1.3
    }
    sunday_value_by_currency = {
        "USD": 3
    }
    today_value_by_currency = {
        "BTC": 11,
        "USD": 54,
        "ETH": 33
    }
    value_by_currency_by_timestamp = {
        friday_timestamp: friday_value_by_currency,     # use exact timestamps for some past timestamps
        sunday_timestamp: sunday_value_by_currency,     # on purpose out of order, value manager is supposed to sort it
        saturday_timestamp + 10: saturday_value_by_currency,
        timestamp: today_value_by_currency,
    }
    historical_portfolio_value_manager.logger = mock.Mock(debug=mock.Mock())
    await historical_portfolio_value_manager.on_new_values(value_by_currency_by_timestamp, save_changes=False)
    assert historical_portfolio_value_manager.get_historical_values("BTC", commons_enums.TimeFrames.ONE_HOUR) == \
        {friday_timestamp: 1.1, saturday_timestamp: 1.3, today_timestamp: 11, today_hour_timestamp: 11}
    # BTC is missing from sunday values
    historical_portfolio_value_manager.logger.debug.assert_called_once()
    historical_portfolio_value_manager.logger.debug.reset_mock()
    assert historical_portfolio_value_manager.get_historical_values("BTC", commons_enums.TimeFrames.ONE_WEEK) == \
        {}
    assert historical_portfolio_value_manager.get_historical_values("USD", commons_enums.TimeFrames.ONE_DAY) == \
        {friday_timestamp: 3010, sunday_timestamp: 3, today_timestamp: 54}
    historical_portfolio_value_manager.logger.debug.assert_called_once()
    historical_portfolio_value_manager.logger.debug.reset_mock()

    late_timestamp = 1648504800  # Monday 28 March 2022 22:00:00 UTC
    late_values = {
        "BTC": 77,
        "USD": 88,
    }
    historical_portfolio_value_manager.portfolio_manager.exchange_manager.exchange.connector.backtesting.\
        time_manager.current_timestamp = late_timestamp
    await historical_portfolio_value_manager.on_new_value(late_timestamp, late_values, save_changes=False)
    assert historical_portfolio_value_manager.get_historical_values("USD", commons_enums.TimeFrames.ONE_DAY) == \
        {friday_timestamp: 3010, sunday_timestamp: 3, today_timestamp: 54}
    historical_portfolio_value_manager.logger.debug.assert_called_once()
    historical_portfolio_value_manager.logger.debug.reset_mock()
    assert historical_portfolio_value_manager.get_historical_values("BTC", commons_enums.TimeFrames.ONE_HOUR) == \
        {friday_timestamp: 1.1, saturday_timestamp: 1.3, today_timestamp: 11, today_hour_timestamp: 11,
         late_timestamp: 77}
    historical_portfolio_value_manager.logger.debug.assert_called_once()
    historical_portfolio_value_manager.logger.debug.reset_mock()

    # add 1h missing value to convertable pairs
    convertable_pair = "BTC/USD"
    price = 3000
    historical_portfolio_value_manager.portfolio_manager.portfolio_value_holder.last_prices_by_trading_pair[
        convertable_pair] = price
    assert historical_portfolio_value_manager.get_historical_values("BTC", commons_enums.TimeFrames.ONE_HOUR) == \
        {friday_timestamp: 1.1, saturday_timestamp: 1.3, sunday_timestamp: 3 / price,
         today_timestamp: 11, today_hour_timestamp: 11, late_timestamp: 77}
    historical_portfolio_value_manager.logger.debug.assert_not_called()

    # with time select
    assert historical_portfolio_value_manager.get_historical_values(
        "BTC", commons_enums.TimeFrames.ONE_HOUR, from_timestamp=sunday_timestamp, to_timestamp=today_timestamp) == \
        {sunday_timestamp: 3 / price, today_timestamp: 11}
    assert historical_portfolio_value_manager.get_historical_values(
        "BTC", commons_enums.TimeFrames.ONE_HOUR, from_timestamp=sunday_timestamp * 2, to_timestamp=today_timestamp) \
        == {}


def _check_historical_value(historical_value, timestamp, value_by_currency):
    assert isinstance(historical_value, personal_data.HistoricalAssetValue)
    assert historical_value.to_dict() == {
        historical_value.TIMESTAMP_KEY: timestamp,
        historical_value.VALUES_KEY: value_by_currency,
    }



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from ccxt import Exchange

import octobot_commons.constants as constants
import octobot_commons.enums as commons_enums
import octobot_trading.enums as trading_enums
from octobot_trading.enums import ExchangeConstantsTickersColumns as Ectc, \
    ExchangeConstantsMarketStatusColumns as Ecmsc
from tests_additional.real_exchanges import get_exchange_manager


class RealExchangeTester:
    # enter exchange name as a class variable here
    EXCHANGE_NAME = None
    EXCHANGE_TYPE = trading_enums.ExchangeTypes.SPOT.value
    SYMBOL = None
    SYMBOL_2 = None
    SYMBOL_3 = None
    # default is 1h, change if necessary
    TIME_FRAME = commons_enums.TimeFrames.ONE_HOUR
    ALLOWED_TIMEFRAMES_WITHOUT_CANDLE = 0

    # Public methods: to be implemented as tests
    # Use await self._[method_name] to get the test request result
    # ex: market_status = await self.get_market_status()

    # unauthenticated API
    async def test_time_frames(self):
        pass

    async def test_get_market_status(self):
        pass

    async def test_get_symbol_prices(self):
        pass

    async def test_get_kline_price(self):
        pass

    async def test_get_order_book(self):
        pass

    async def test_get_recent_trades(self):
        pass

    async def test_get_price_ticker(self):
        pass

    async def test_get_all_currencies_price_ticker(self):
        pass

    # authenticated API
    # TODO
    # async def test_get_balance(self):
    #     pass
    #
    # async def test_get_order(self):
    #     pass
    #
    # async def test_get_all_orders(self):
    #     pass
    #
    # async def test_get_open_orders(self):
    #     pass
    #
    # async def test_get_closed_orders(self):
    #     pass
    #
    # async def test_get_my_recent_trades(self):
    #     pass
    #
    # async def test_cancel_order(self):
    #     pass
    #
    # async def test_create_order(self):
    #     pass
    
    def get_config(self):
        return {
            constants.CONFIG_EXCHANGES: {
                self.EXCHANGE_NAME: {
                    constants.CONFIG_EXCHANGE_TYPE: self.EXCHANGE_TYPE
                }
            }
        }

    async def time_frames(self):
        async with get_exchange_manager(self.EXCHANGE_NAME, self.get_config()) as exchange_manager:
            return exchange_manager.exchange.time_frames

    async def get_market_statuses(self):
        # return 2 different market status with different traded pairs to reduce possible
        # side effects using only one pair.
        async with get_exchange_manager(self.EXCHANGE_NAME, self.get_config()) as exchange_manager:
            return exchange_manager.exchange.get_market_status(self.SYMBOL), \
                   exchange_manager.exchange.get_market_status(self.SYMBOL_2), \
                   exchange_manager.exchange.get_market_status(self.SYMBOL_3)

    async def get_symbol_prices(self, limit=None, **kwargs):
        async with get_exchange_manager(self.EXCHANGE_NAME, self.get_config()) as exchange_manager:
            return await exchange_manager.exchange.get_symbol_prices(self.SYMBOL, self.TIME_FRAME,
                                                                     limit=limit, **kwargs)

    async def get_kline_price(self, **kwargs):
        async with get_exchange_manager(self.EXCHANGE_NAME, self.get_config()) as exchange_manager:
            return await exchange_manager.exchange.get_kline_price(self.SYMBOL, self.TIME_FRAME, **kwargs)

    async def get_order_book(self, **kwargs):
        async with get_exchange_manager(self.EXCHANGE_NAME, self.get_config()) as exchange_manager:
            return await exchange_manager.exchange.get_order_book(self.SYMBOL, **kwargs)

    async def get_recent_trades(self, limit=50):
        async with get_exchange_manager(self.EXCHANGE_NAME, self.get_config()) as exchange_manager:
            return await exchange_manager.exchange.get_recent_trades(self.SYMBOL, limit=limit)

    async def get_price_ticker(self):
        async with get_exchange_manager(self.EXCHANGE_NAME, self.get_config()) as exchange_manager:
            return await exchange_manager.exchange.get_price_ticker(self.SYMBOL)

    async def get_all_currencies_price_ticker(self, **kwargs):
        async with get_exchange_manager(self.EXCHANGE_NAME, self.get_config()) as exchange_manager:
            return await exchange_manager.exchange.get_all_currencies_price_ticker(**kwargs)

    def get_allowed_time_delta(self):
        return (self.ALLOWED_TIMEFRAMES_WITHOUT_CANDLE + 1) * \
               commons_enums.TimeFramesMinutes[self.TIME_FRAME] * \
               constants.MINUTE_TO_SECONDS * constants.MSECONDS_TO_SECONDS * 1.3

    @staticmethod
    def get_time():
        return Exchange.milliseconds()

    @staticmethod
    def ensure_elements_order(elements, sort_key, reverse=False):
        assert sorted(elements, key=lambda x: x[sort_key], reverse=reverse) == elements

    def check_market_status_limits(self, market_status,
                                   normal_price_max=10000, normal_price_min=1e-06,
                                   normal_cost_max=10000, normal_cost_min=1e-06,
                                   low_price_max=1e-07, low_price_min=1e-09,
                                   low_cost_max=1e-03, low_cost_min=1e-06,
                                   expect_invalid_price_limit_values=False,
                                   expect_inferior_or_equal_price_and_cost=False,
                                   enable_price_and_cost_comparison=True):
        min_price = market_status[Ecmsc.LIMITS.value][Ecmsc.LIMITS_PRICE.value][Ecmsc.LIMITS_PRICE_MIN.value]
        min_cost = market_status[Ecmsc.LIMITS.value][Ecmsc.LIMITS_COST.value][Ecmsc.LIMITS_COST_MIN.value]
        has_price_limit_value = min_price not in (None, 0)
        has_cost_limit_value = min_cost not in (None, 0)
        has_limit_value = has_price_limit_value and has_cost_limit_value
        if (not has_limit_value) and expect_invalid_price_limit_values:
            raise AssertionError("No price and limit value does not mean invalid values")
        if not expect_invalid_price_limit_values and market_status[Ecmsc.SYMBOL.value] == self.SYMBOL_3:
            # if these test are not passing, it means that limits are invalid
            # (limits should be much lower for SYMBOL_3 which is the low price pair, ex: XRP/BTC)
            # => set expect_invalid_price_limit_values to True in call and
            # remove price limit in exchange tentacle market status fixer if this is the case
            assert (not has_price_limit_value) or low_price_max >= min_price >= low_price_min
            assert (not has_cost_limit_value) or low_cost_max >= min_cost >= low_cost_min
        else:
            assert (not has_price_limit_value) or normal_price_max >= min_price >= normal_price_min
            assert (not has_cost_limit_value) or normal_cost_max >= min_cost >= normal_cost_min
        if enable_price_and_cost_comparison and has_limit_value:
            # Consistency here is not required by OctoBot. Fix in tentacles if consistency
            # in price/cost comparison becomes required and min_price <= min_cost is false without a good reason
            if expect_inferior_or_equal_price_and_cost:
                assert min_price >= min_cost
            else:
                assert min_price <= min_cost

    @staticmethod
    def check_ticker_typing(ticker, check_open=True, check_high=True, check_low=True,
                            check_close=True, check_base_volume=True, check_timestamp=True):
        if check_open:
            assert isinstance(ticker[Ectc.OPEN.value], (float, int))
        if check_high:
            assert isinstance(ticker[Ectc.HIGH.value], (float, int))
        if check_low:
            assert isinstance(ticker[Ectc.LOW.value], (float, int))
        if check_close:
            assert isinstance(ticker[Ectc.CLOSE.value], (float, int))
        if check_base_volume:
            assert isinstance(ticker[Ectc.BASE_VOLUME.value], (float, int))
        if check_timestamp:
            assert isinstance(ticker[Ectc.TIMESTAMP.value], (float, int))



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import math

import octobot_commons.logging as logging
from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc
from octobot_trading.enums import ExchangeConstantsMarketStatusInfoColumns as Ecmsic


def is_ms_valid(value, zero_valid=False):
    if isinstance(value, str):
        try:
            value = float(value)
        except ValueError:
            return False
    return value is not None and value is not math.nan and (value >= 0 if zero_valid else value > 0)


def check_market_status_limits(market_limit):
    return all([check_market_status_values(market_limit[key].values())
                for key in market_limit])


def check_market_status_values(values, zero_valid=False):
    return all([is_ms_valid(value, zero_valid=zero_valid) for value in values])


def get_markets_limit(market_limit):
    return market_limit[Ecmsc.LIMITS_COST.value] if Ecmsc.LIMITS_COST.value in market_limit else None, \
           market_limit[Ecmsc.LIMITS_PRICE.value] if Ecmsc.LIMITS_PRICE.value in market_limit else None, \
           market_limit[Ecmsc.LIMITS_AMOUNT.value] if Ecmsc.LIMITS_AMOUNT.value in market_limit else None


def calculate_amounts(market_limit):
    limit_cost, limit_price, limit_amount = get_markets_limit(market_limit)

    if not is_ms_valid(limit_amount[Ecmsc.LIMITS_AMOUNT_MAX.value]) and \
        Ecmsc.LIMITS_COST_MAX.value in limit_cost and Ecmsc.LIMITS_PRICE_MAX.value in limit_price:
        if is_ms_valid(limit_cost[Ecmsc.LIMITS_COST_MAX.value]) \
                and is_ms_valid(limit_price[Ecmsc.LIMITS_PRICE_MAX.value]) \
                and limit_price[Ecmsc.LIMITS_PRICE_MAX.value] > 0:
            limit_amount[Ecmsc.LIMITS_AMOUNT_MAX.value] = limit_cost[Ecmsc.LIMITS_COST_MAX.value] / \
                                                          limit_price[Ecmsc.LIMITS_PRICE_MAX.value]

    if not is_ms_valid(limit_amount[Ecmsc.LIMITS_AMOUNT_MIN.value]) and \
        Ecmsc.LIMITS_COST_MIN.value in limit_cost and Ecmsc.LIMITS_PRICE_MIN.value in limit_price:
        if is_ms_valid(limit_cost[Ecmsc.LIMITS_COST_MIN.value]) \
                and is_ms_valid(limit_price[Ecmsc.LIMITS_PRICE_MIN.value]) \
                and limit_price[Ecmsc.LIMITS_PRICE_MIN.value] > 0:
            limit_amount[Ecmsc.LIMITS_AMOUNT_MIN.value] = limit_cost[Ecmsc.LIMITS_COST_MIN.value] / \
                                                          limit_price[Ecmsc.LIMITS_PRICE_MIN.value]


def calculate_costs(market_limit):
    limit_cost, limit_price, limit_amount = get_markets_limit(market_limit)

    if not is_ms_valid(limit_cost[Ecmsc.LIMITS_COST_MAX.value]) and \
        Ecmsc.LIMITS_AMOUNT_MAX.value in limit_amount and Ecmsc.LIMITS_PRICE_MAX.value in limit_price:
        if is_ms_valid(limit_amount[Ecmsc.LIMITS_AMOUNT_MAX.value]) \
                and is_ms_valid(limit_price[Ecmsc.LIMITS_PRICE_MAX.value]):
            limit_cost[Ecmsc.LIMITS_COST_MAX.value] = limit_amount[Ecmsc.LIMITS_AMOUNT_MAX.value] * \
                                                      limit_price[Ecmsc.LIMITS_PRICE_MAX.value]

    if not is_ms_valid(limit_cost[Ecmsc.LIMITS_COST_MIN.value]) and \
        Ecmsc.LIMITS_AMOUNT_MIN.value in limit_amount and Ecmsc.LIMITS_PRICE_MIN.value in limit_price:
        if is_ms_valid(limit_amount[Ecmsc.LIMITS_AMOUNT_MIN.value]) \
                and is_ms_valid(limit_price[Ecmsc.LIMITS_PRICE_MIN.value]):
            limit_cost[Ecmsc.LIMITS_COST_MIN.value] = limit_amount[Ecmsc.LIMITS_AMOUNT_MIN.value] * \
                                                      limit_price[Ecmsc.LIMITS_PRICE_MIN.value]


def calculate_prices(market_limit):
    limit_cost, limit_price, limit_amount = get_markets_limit(market_limit)

    if not is_ms_valid(limit_price[Ecmsc.LIMITS_PRICE_MAX.value]) and \
        Ecmsc.LIMITS_COST_MAX.value in limit_cost and Ecmsc.LIMITS_AMOUNT_MAX.value in limit_amount:
        if is_ms_valid(limit_cost[Ecmsc.LIMITS_COST_MAX.value]) \
                and is_ms_valid(limit_amount[Ecmsc.LIMITS_AMOUNT_MAX.value]) \
                and limit_amount[Ecmsc.LIMITS_AMOUNT_MAX.value] > 0:
            limit_price[Ecmsc.LIMITS_PRICE_MAX.value] = limit_cost[Ecmsc.LIMITS_COST_MAX.value] / \
                                                        limit_amount[Ecmsc.LIMITS_AMOUNT_MAX.value]

    if not is_ms_valid(limit_price[Ecmsc.LIMITS_PRICE_MIN.value]) and \
            Ecmsc.LIMITS_COST_MIN.value in limit_cost and Ecmsc.LIMITS_AMOUNT_MIN.value in limit_amount:
        if is_ms_valid(limit_cost[Ecmsc.LIMITS_COST_MIN.value]) \
                and is_ms_valid(limit_amount[Ecmsc.LIMITS_AMOUNT_MIN.value]) \
                and limit_amount[Ecmsc.LIMITS_AMOUNT_MIN.value] > 0:
            limit_price[Ecmsc.LIMITS_PRICE_MIN.value] = limit_cost[Ecmsc.LIMITS_COST_MIN.value] / \
                                                        limit_amount[Ecmsc.LIMITS_AMOUNT_MIN.value]


def fix_market_status_limits_from_current_data(market_limit):
    # calculate cost
    if not (check_market_status_values(market_limit[Ecmsc.LIMITS_COST.value].values())):
        calculate_costs(market_limit)

    # calculate amounts
    if not (check_market_status_values(market_limit[Ecmsc.LIMITS_AMOUNT.value].values())):
        calculate_amounts(market_limit)

    # calculate prices
    if not (check_market_status_values(market_limit[Ecmsc.LIMITS_PRICE.value].values())):
        calculate_prices(market_limit)

    if not is_ms_valid(
            market_limit[Ecmsc.LIMITS_COST.value][Ecmsc.LIMITS_COST_MIN.value]):
        market_limit[Ecmsc.LIMITS_COST.value][Ecmsc.LIMITS_COST_MIN.value] = 0


class ExchangeMarketStatusFixer:
    LIMIT_PRICE_MULTIPLIER = 1000
    LIMIT_COST_MULTIPLIER = 1

    # calculated from popular exchanges
    LIMIT_AMOUNT_MAX_SUP_ATTENUATION = 6  # when log(price, 10) >= 0
    LIMIT_AMOUNT_MAX_MINUS_3_ATTENUATION = 1  # when log(price, 10) > -3
    LIMIT_AMOUNT_MIN_ATTENUATION = 3  # when log(price, 10) < 0
    LIMIT_AMOUNT_MIN_SUP_ATTENUATION = 1  # when log(price, 10) >= 0

    """
    Utility class that performs exchange_self.market_status fixes
    """

    def __init__(self, market_status, price_example=None):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.market_status = market_status
        self.price_example = price_example

        if Ecmsc.INFO.value in self.market_status:
            self.market_status_specific = self.market_status[Ecmsc.INFO.value]
        else:
            self.market_status_specific = None

        self._fix_typing()
        self._fix_market_status_precision()
        self._fix_market_status_limits()

    def _convert_values_to_float(self, element, parent_keys, key_whitelist):
        try:
            for parent_key in parent_keys:
                element = element[parent_key]
        except KeyError:
            # no parent_key in element, nothing to convert
            return
        for key, val in element.items():
            if key in key_whitelist and isinstance(val, str):
                try:
                    element[key] = float(val)
                except ValueError:
                    self.logger.debug(f"Impossible to convert {val} to float in {key} market status. "
                                      f"Full market status: {self.market_status}")

    def _fix_typing(self):
        self._convert_values_to_float(
            self.market_status,
            [Ecmsc.PRECISION.value, ],
            [Ecmsc.PRECISION_AMOUNT.value, Ecmsc.PRECISION_COST.value, Ecmsc.PRECISION_PRICE.value],
        )
        self._convert_values_to_float(
            self.market_status,
            [Ecmsc.LIMITS.value, Ecmsc.LIMITS_COST.value],
            [Ecmsc.LIMITS_COST_MAX.value, Ecmsc.LIMITS_COST_MIN.value],
        )
        self._convert_values_to_float(
            self.market_status,
            [Ecmsc.LIMITS.value, Ecmsc.LIMITS_AMOUNT.value],
            [Ecmsc.LIMITS_AMOUNT_MAX.value, Ecmsc.LIMITS_AMOUNT_MIN.value],
        )
        self._convert_values_to_float(
            self.market_status,
            [Ecmsc.LIMITS.value, Ecmsc.LIMITS_PRICE.value],
            [Ecmsc.LIMITS_PRICE_MAX.value, Ecmsc.LIMITS_PRICE_MIN.value],
        )

    def _fix_market_status_precision(self):
        if Ecmsc.PRECISION.value not in self.market_status:
            self.market_status[Ecmsc.PRECISION.value] = {
                Ecmsc.PRECISION_AMOUNT.value: None,
                Ecmsc.PRECISION_COST.value: None,
                Ecmsc.PRECISION_PRICE.value: None,
            }

        market_precision = self.market_status[Ecmsc.PRECISION.value]

        if not check_market_status_values(market_precision.values(), zero_valid=True):
            if self.price_example is not None:
                self._fix_market_status_precision_with_price()

            elif self.market_status_specific:
                self._fix_market_status_precision_with_specific()

    def _fix_market_status_limits(self):
        if Ecmsc.LIMITS.value not in self.market_status:
            self.market_status[Ecmsc.LIMITS.value] = {}

        market_limit = self.market_status[Ecmsc.LIMITS.value]

        if Ecmsc.LIMITS_COST.value not in market_limit:
            market_limit[Ecmsc.LIMITS_COST.value] = {
                Ecmsc.LIMITS_COST_MAX.value: None,
                Ecmsc.LIMITS_COST_MIN.value: None
            }

        if Ecmsc.LIMITS_AMOUNT.value not in market_limit:
            market_limit[Ecmsc.LIMITS_AMOUNT.value] = {
                Ecmsc.LIMITS_AMOUNT_MAX.value: None,
                Ecmsc.LIMITS_AMOUNT_MIN.value: None
            }

        if Ecmsc.LIMITS_PRICE.value not in market_limit:
            market_limit[Ecmsc.LIMITS_PRICE.value] = {
                Ecmsc.LIMITS_PRICE_MAX.value: None,
                Ecmsc.LIMITS_PRICE_MIN.value: None
            }

        # if some data is missing
        if not check_market_status_limits(market_limit):
            fix_market_status_limits_from_current_data(market_limit)

            if self.market_status_specific and not check_market_status_limits(market_limit):
                self._fix_market_status_limits_with_specific()

            if self.price_example is not None and not check_market_status_limits(market_limit):
                self._fix_market_status_limits_with_price()

    def _calculate_amount(self):
        amount_log_price = math.log(self.price_example, 10)

        if amount_log_price >= 0:
            amount_min = 10 ** (self.LIMIT_AMOUNT_MIN_SUP_ATTENUATION - amount_log_price)
            amount_max = 10 ** (self.LIMIT_AMOUNT_MAX_SUP_ATTENUATION - amount_log_price)
        else:
            amount_min = 10 ** -(amount_log_price + self.LIMIT_AMOUNT_MIN_ATTENUATION)
            amount_max = 10 ** (-amount_log_price + self.LIMIT_AMOUNT_MAX_MINUS_3_ATTENUATION)

        return amount_min, amount_max

    def _fix_market_status_limits_with_price(self):
        # self.LIMIT_AMOUNT_MULTIPLIER
        amount_min, amount_max = self._calculate_amount()
        price_max = self.price_example * self.LIMIT_PRICE_MULTIPLIER
        price_min = self.price_example / self.LIMIT_PRICE_MULTIPLIER

        self.market_status[Ecmsc.LIMITS.value] = {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: amount_min,
                Ecmsc.LIMITS_AMOUNT_MAX.value: amount_max,
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: price_min,
                Ecmsc.LIMITS_PRICE_MAX.value: price_max,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: price_min * amount_min,
                Ecmsc.LIMITS_COST_MAX.value: price_max * amount_max,
            }
        }

    def _get_price_precision(self):
        return -decimal.Decimal(f"{self.price_example}").as_tuple().exponent

    def _fix_market_status_precision_with_price(self):
        precision = self._get_price_precision()
        self.market_status[Ecmsc.PRECISION.value] = {
            Ecmsc.PRECISION_AMOUNT.value: precision,
            Ecmsc.PRECISION_COST.value: precision,
            Ecmsc.PRECISION_PRICE.value: precision,
        }

    def _fix_market_status_precision_with_specific(self):
        # binance specific
        pass  # nothing for binance

    def _fix_market_status_limits_with_specific(self):
        market_limit = self.market_status[Ecmsc.LIMITS.value]

        try:
            # binance specific
            if Ecmsic.FILTERS.value in self.market_status_specific:
                filters = self.market_status_specific[Ecmsic.FILTERS.value]

                for filter_dict in filters:
                    if Ecmsic.FILTER_TYPE.value in filter_dict:
                        if filter_dict[Ecmsic.FILTER_TYPE.value] == Ecmsic.PRICE_FILTER.value:
                            if is_ms_valid(float(filter_dict[Ecmsic.MAX_PRICE.value])) and \
                                    not is_ms_valid(
                                        market_limit[Ecmsc.LIMITS_PRICE.value][Ecmsc.LIMITS_PRICE_MAX.value]):
                                market_limit[Ecmsc.LIMITS_PRICE.value][Ecmsc.LIMITS_PRICE_MAX.value] = \
                                    float(filter_dict[Ecmsic.MAX_PRICE.value])

                            if is_ms_valid(float(filter_dict[Ecmsic.MIN_PRICE.value])) and \
                                    not is_ms_valid(
                                        market_limit[Ecmsc.LIMITS_PRICE.value][Ecmsc.LIMITS_PRICE_MIN.value]):
                                market_limit[Ecmsc.LIMITS_PRICE.value][Ecmsc.LIMITS_PRICE_MIN.value] = \
                                    float(filter_dict[Ecmsic.MIN_PRICE.value])
                        elif filter_dict[Ecmsic.FILTER_TYPE.value] == Ecmsic.LOT_SIZE.value:
                            if is_ms_valid(float(filter_dict[Ecmsic.MAX_QTY.value])) and \
                                    not is_ms_valid(
                                        market_limit[Ecmsc.LIMITS_AMOUNT.value][Ecmsc.LIMITS_AMOUNT_MAX.value]):
                                market_limit[Ecmsc.LIMITS_AMOUNT.value][Ecmsc.LIMITS_AMOUNT_MAX.value] = \
                                    float(filter_dict[Ecmsic.MAX_QTY.value])

                            if is_ms_valid(float(filter_dict[Ecmsic.MIN_QTY.value])) and \
                                    not is_ms_valid(
                                        market_limit[Ecmsc.LIMITS_AMOUNT.value][Ecmsc.LIMITS_AMOUNT_MIN.value]):
                                market_limit[Ecmsc.LIMITS_AMOUNT.value][Ecmsc.LIMITS_AMOUNT_MIN.value] = \
                                    float(filter_dict[Ecmsic.MIN_QTY.value])
                calculate_costs(market_limit)
        except Exception:
            pass
